/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Description
    Splits mesh by making internal faces external. Uses attachDetach.

    Generates a meshModifier of the form:

    Splitter
    {
        type                       attachDetach;
        faceZoneName               membraneFaces;
        masterPatchName            masterPatch;
        slavePatchName             slavePatch;
        triggerTimes               runTime.value();
    }

    so will detach at the current time and split all faces in membraneFaces
    into masterPatch and slavePatch (which have to be present but of 0 size)

    Preserves additional faceSets. The faceSet that was used to split the mesh is erased.
    Assumes that the cells are not renumbered

 Based on ICE Revision: $Id: splitMeshWithSets.C 7425 2007-04-25 09:20:34Z bgschaid $

Compiled on OF-2.1.x by Hisham El Safti
\*---------------------------------------------------------------------------*/

#include "argList.H"
#include "polyMesh.H"
#include "Time.H"
#include "polyTopoChange.H"
#include "mapPolyMesh.H"
#include "faceSet.H"
#include "attachDetach.H"
#include "attachPolyTopoChanger.H"
#include "regionSide.H"
#include "primitiveFacePatch.H"

using namespace Foam;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

// Find edge between points v0 and v1.
label findEdge(const primitiveMesh& mesh, const label v0, const label v1)
{
    const labelList& pEdges = mesh.pointEdges()[v0];

    forAll(pEdges, pEdgeI)
    {
        label edgeI = pEdges[pEdgeI];

        const edge& e = mesh.edges()[edgeI];

        if (e.otherVertex(v0) == v1)
        {
            return edgeI;
        }
    }

    FatalErrorIn
    (
        "findEdge(const primitiveMesh&, const label, const label)"
    )   << "Cannot find edge between mesh points " << v0 << " and " << v1
        << abort(FatalError);

    return -1;
}


// Checks whether patch present
void checkPatch(const polyBoundaryMesh& bMesh, const word& name)
{
    const label patchI = bMesh.findPatchID(name);

    if (patchI == -1)
    {
        FatalErrorIn("checkPatch(const polyBoundaryMesh&, const word&)")
            << "Cannot find patch " << name << nl
            << "It should be present but of zero size" << endl
            << "Valid patches are " << bMesh.names()
            << exit(FatalError);
    }

    if (bMesh[patchI].size())
    {
        FatalErrorIn("checkPatch(const polyBoundaryMesh&, const word&)")
            << "Patch " << name << " is present but non-zero size"
            << exit(FatalError);
    }
}


wordList getFaceSets(const polyMesh &m) {
    DynamicList<word> sets;
    fileName setDir=m.rootPath()/m.caseName()/"constant"/polyMesh::meshSubDir/"sets";

    fileNameList files=readDir(setDir);

    forAll(files,i) {
        IOobject f
            (
                files[i],
                "constant",
                polyMesh::meshSubDir/"sets",
                m,
                IOobject::MUST_READ,
                IOobject::NO_WRITE
                
            );
        f.headerOk();
        if(f.headerClassName()=="faceSet") {
            sets.append(files[i]);
        }
    }

    sets.shrink();

    return sets;
}


// Main program:

int main(int argc, char *argv[])
{
    argList::noParallel();
#   include "addOverwriteOption.H"

    argList::validArgs.append("faceSet");
    argList::validArgs.append("masterPatch");
    argList::validArgs.append("slavePatch");

#   include "setRootCase.H"
#   include "createTime.H"
    runTime.functionObjects().off();
#   include "createPolyMesh.H"
    const word oldInstance = mesh.pointsInstance();

    const word setName = args[1];
    const word masterPatch = args[2];
    const word slavePatch = args[3];
    const bool overwrite = args.optionFound("overwrite");

    wordList theFaceSets=getFaceSets(mesh);
    //   Info << "Face sets: " << theFaceSets << endl;

    // List of faces to split
    faceSet facesSet(mesh, setName);

    Info<< "Read " << facesSet.size() << " faces to split" << endl << endl;


    // Convert into labelList and check

    labelList faces(facesSet.toc());

    forAll(faces, i)
    {
        if (!mesh.isInternalFace(faces[i]))
        {
            FatalErrorIn(args.executable())
            << "Face " << faces[i] << " in faceSet " << setName
            << " is not an internal face."
            << exit(FatalError);
        }
    }


    // Check for empty master and slave patches
    checkPatch(mesh.boundaryMesh(), masterPatch);
    checkPatch(mesh.boundaryMesh(), slavePatch);


    //
    // Find 'side' of all faces on splitregion. Uses regionSide which needs
    // set of edges on side of this region. Use PrimitivePatch to find these.
    //

    // Addressing on faces only in mesh vertices.
    primitiveFacePatch fPatch
    (
        faceList
        (
            UIndirectList<face>
            (
                mesh.faces(),
                faces
            )
        ),
        mesh.points()
    );

    const labelList& meshPoints = fPatch.meshPoints();

    // Mark all fence edges : edges on boundary of fPatch but not on boundary
    // of polyMesh
    labelHashSet fenceEdges(fPatch.size());

    const labelListList& allEdgeFaces = fPatch.edgeFaces();

    forAll(allEdgeFaces, patchEdgeI)
    {
        if (allEdgeFaces[patchEdgeI].size() == 1)
        {
            const edge& e = fPatch.edges()[patchEdgeI];

            label edgeI =
                findEdge
                (
                    mesh,
                    meshPoints[e.start()],
                    meshPoints[e.end()]
                );

            fenceEdges.insert(edgeI);
        }
    }

    // Find sides reachable from 0th face of faceSet
    label startFaceI = faces[0];

    regionSide regionInfo
    (
        mesh,
        facesSet,
        fenceEdges,
        mesh.faceOwner()[startFaceI],
        startFaceI
    );

    // Determine flip state for all faces in faceSet
    boolList zoneFlip(faces.size());

    forAll(faces, i)
    {
        zoneFlip[i] = !regionInfo.sideOwner().found(faces[i]);
    }


    

    /* label nrPointZones=mesh.pointZones().size();
    label nrCellZones=mesh.cellZones().size();
    label nrFaceZones=mesh.faceZones().size();

    bool zonesExist=false;

    if(nrPointZones>0 || nrCellZones >0 || nrFaceZones>0) {
        Info << " Allready zones existing in the mesh " << endl;

        if(nrFaceZones>0) {
            FatalErrorIn(args.executable())
                << " Already " << nrFaceZones << " faceZones existing"
                    << exit(FatalError);            
        }

        zonesExist=true;
        }*/

    
 // Create and add face zones and mesh modifiers
    List<pointZone*> pz(0);
    List<faceZone*> fz(1);
    List<cellZone*> cz(0);

    fz[0] =
        new faceZone
        (
            "membraneFaces",
            faces,
            zoneFlip,
            0,
            mesh.faceZones()
        );

    Info<< "Adding point and face zones" << endl;
    mesh.addZones(pz, fz, cz);

    HashTable<faceSet*,word> theSets;

    HashTable<labelPairList*,word> cellInfo;

    // Get information about the cells on each side of the faceSet
    {
        forAll(theFaceSets,i) {
            if(theFaceSets[i]!=setName) {
                word name=theFaceSets[i];
                Info << "Adding set for faceSet " << name << endl;
                theSets.insert(name,new faceSet(mesh,name));
                faceSet &thisSet=*theSets[name];
                labelList theFaces(thisSet.toc());
                cellInfo.insert(name,new labelPairList(theFaces.size()));
                labelPairList &faceCells=*cellInfo[name];
                const labelList& own=mesh.faceOwner();
                const labelList& neigh=mesh.faceNeighbour();
                forAll(theFaces,faceI) {
                    label ind=theFaces[faceI];
                    faceCells[faceI].first()=own[ind];
                    faceCells[faceI].second()=neigh[ind];
                }
            }
        }
    }

    /*    if(!zonesExist) {

        Info<< "Adding point and face zones" << endl;

        // Create and add face zones and mesh modifiers
        List<pointZone*> pz(0);
        List<faceZone*> fz(1);
        fz[0] =
            new faceZone
            (
                "membraneFaces",
                faces,
                zoneFlip,
                0,
                mesh.faceZones()
            );

        List<cellZone*> cz(0);

        mesh.addZones(pz, fz, cz);

    } else {

        Info << "Adding the faceZone" << endl;
        
        faceZone* fc;
        fc =
            new faceZone
            (
                "membraneFaces",
                faces,
                zoneFlip,
                0,
                mesh.faceZones()
            );

        mesh.faceZones().setSize(1);
        mesh.faceZones().set(0,fc);
        }*/

    attachPolyTopoChanger splitter(mesh);
    splitter.setSize(1);

    // Add the sliding interface mesh modifier to start working at current
    // time
    splitter.set
    (
        0,
        new attachDetach
        (
            "Splitter",
            0,
            splitter,
            "membraneFaces",
            masterPatch,
            slavePatch,
            scalarField(1, runTime.value())
        )
    );

    Info<< nl << "Constructed topologyModifier:" << endl;
    splitter[0].writeDict(Info);

    if (!overwrite)
    {
        runTime++;
    }

    splitter.attach();

    // Reconstruct the faceSets from the cellNumbers 
    {
        const labelList& own=mesh.faceOwner();
        const labelList& neigh=mesh.faceNeighbour();
        const cellList &cells=mesh.cells();

        wordList names=theSets.toc();
        forAll(names,i) {
            word name=names[i];

            labelPairList &faceCells=*cellInfo[name];
            faceSet &thisSet=*theSets[name];
            labelList theFaces(thisSet.toc());

            Info << "Resetting faceSet " << name << endl;
            thisSet.clear();

            forAll(faceCells,j) {
                label cellA=faceCells[j].first();
                label cellB=faceCells[j].second();
                const cell &c=cells[cellA];
                label foundFace=-1;

                forAll(c,k) {
                    label fI=c[k];
                    if( (own[fI]==cellA && neigh[fI]==cellB) || (own[fI]==cellB && neigh[fI]==cellA) ) {
                        foundFace=fI;
                        break;
                    }
                }
                if(foundFace<0) {
                    //                     FatalErrorIn(args.executable())
                    Info    << " No matching face between cells " << faceCells[j] << " in set " << name << "found!";
                    //                             << exit(FatalError);
                } else {
                    thisSet.insert(foundFace);
                }
            }
        }
    }

    if (overwrite)
    {
        mesh.setInstance(oldInstance);
    }
    else
    {
        mesh.setInstance(runTime.timeName());
    }

    //mesh.faceZones().clear();
    
    /*if(!zonesExist) {
        mesh.faceZones().writeOpt()=IOobject::NO_WRITE;
        mesh.cellZones().writeOpt()=IOobject::NO_WRITE;
        mesh.pointZones().writeOpt()=IOobject::NO_WRITE;
        }*/

    splitter.clear();
    splitter.writeOpt()=IOobject::NO_WRITE;

    Info<< "Writing mesh to " << runTime.timeName() << endl;
    if (!mesh.write())
    {
        FatalErrorIn(args.executable())
            << "Failed writing polyMesh."
            << exit(FatalError);
    }

    Info << "Erasing face set " << setName << endl;
    facesSet.clear();
    facesSet.write();

    // Write the reconstructed faceSets
    {
        wordList names=theSets.toc();

        forAll(names,i) {
            word name=names[i];
            Info << "Writing faceSet " << name << " ( " << theSets[name]->size() << " ) " << endl;
            theSets[name]->write();
        }
    }

    Info<< nl << "end" << endl;
    return 0;
}


// ************************************************************************* //
